/*
 
 Copyright 1996 by J.D. Koftinoff Software Ltd, 
 under the KPL (Koftinoff Public License)
 
 THE NO-NONSENSE KOFTINOFF PUBLIC LICENSE (KPL)

 
 This software is provided free of charge.  You use this software only under 
 the following conditions:
 
    * You agree that you use this software at your own risk
    * You agree that the copyright of this software is held 
      by J.D. Koftinoff Software, Ltd.
 
 If you agree to these conditions, you may modify and use the software in 
 commercial as well as free projects.  However, If you use this software in a 
 commercial project, you MUST send a neat postcard to me at:
 
    Jeff Koftinoff
    J.D. Koftinoff Software, Ltd.
    #5 - 1131 Burnaby St.
    Vancouver, B.C.
    Canada
    V6E-1P3
 
    email: jeffk@jdkoftinoff.com
 
 Have Fun!
 
 P.S. Send me a postcard anyways please! I'll send one back!
 */


#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/wait.h>
#include "jkcthread.h"

#if 0
#define DBG(XXX)     printf XXX
#else
#define DBG(XXX)
#endif


#define CSIGNAL         0x000000ff      /* signal mask to be sent at exit */
#define CLONE_VM        0x00000100      /* set if VM shared between processes */
#define CLONE_FS        0x00000200      /* set if fs info shared between processes */
#define CLONE_FILES     0x00000400      /* set if open files shared between processes */
#define CLONE_SIGHAND   0x00000800      /* set if signal handlers shared */


/* libc memory management routines */
extern void * __libc_malloc( size_t );
extern void * __libc_calloc( size_t, size_t );
extern void * __libc_realloc( void *, size_t );
extern void __libc_free(void * );


/* internal routines */
void _jkthread_SIGCHLD (int signum);
void _jkthread_exit(void);
void _jkthread_SIGINT(int);
void _jkthread_SIGTERM(int);
int _linux_thread_create(void (*fn)(void *), void *data, int stacksz);
void _jkthread_starter( void *p );


/* global data for keeping track of threads and semaphores */

jkthreadinfo _jkthreadinfos[JKMAX_THREADS];
jkthreadinfo _jkdummythread;
jksem _jksems[JKMAX_SEMS];
jksem _jksyssem;
jksem _jklibc_sem;
int _jkthread_inited = 0;

jkthreadinfo *_jkthread_kludge[32768];

int *__errno_location(void)
{
    return &_jkthread_kludge[ getpid() ]->thr_errno;	
}

int *__h_errno_location(void)
{
    return &_jkthread_kludge[ getpid() ]->thr_h_errno;		
}

int *jkthread_getlocal(void)
{
    return _jkthread_kludge[ getpid() ]->thr_local;    	
}

void * malloc( size_t sz)
{
	void * ret;
	
	if( !_jkthread_inited )
	    jkthread_init();
	
    jkrsem_get( &_jklibc_sem );

	ret = __libc_malloc( sz );

	jkrsem_release( &_jklibc_sem );
	return ret;
}


void * calloc( size_t a, size_t b )
{
	void * ret;

	if( !_jkthread_inited )
	    jkthread_init();
	
    jkrsem_get( &_jklibc_sem );

	ret = __libc_calloc( a,b );

	jkrsem_release( &_jklibc_sem );
	return ret;
}

void * realloc( void * p, size_t a )
{
	void * ret;

	if( !_jkthread_inited )
	    jkthread_init();
	
    jkrsem_get( &_jklibc_sem );

	ret = __libc_realloc( p, a );

	jkrsem_release( &_jklibc_sem );
	return ret;
}

void free(void * p)
{
	if( !_jkthread_inited )
	    jkthread_init();
	
    jkrsem_get( &_jklibc_sem );

	__libc_free( p );

	jkrsem_release( &_jklibc_sem );
}


unsigned int sleep(unsigned int a)
{
    struct timeval tv;
	
	if( !_jkthread_inited )
	    jkthread_init();
	
	tv.tv_sec = a;
	tv.tv_usec=0;

	do
	{		
    	select(0,0,0,0,&tv );
	} while( tv.tv_sec!=0 || tv.tv_usec!=0 );

	return 0;
}


void jkthread_init(void)
{
    int i;
	
	for( i=0; i<JKMAX_THREADS; ++i )
	{
	    _jkthreadinfos[i].thr_id=0;
	    _jkthreadinfos[i].thr_errno=0;
	    _jkthreadinfos[i].thr_h_errno=0;		

	}

	for( i=0; i<JKMAX_SEMS; ++i )
	{
	    _jksems[i].sem_id=0;
		_jksems[i].usage=0;
		_jksems[i].owner=0;
	}

	for( i=0; i<32768; ++i )
	{
		/* dummythread is required so errno works temporarily while we wait for
		 * kludge to be updated after thread creation
		 */
		
	    _jkthread_kludge[i] = &_jkdummythread;	
	}
	
	_jkthread_inited=1;	
	
	/* _jksyssem is the semaphore for our thread managment routines */

	jksem_init( &_jksyssem );
	
	/* _jklibc_sem is the semaphore for the non-reentrant libc stuff */
	
	jksem_init( &_jklibc_sem );	
		
	/* The first thread is the main thread */

	_jkthreadinfos[0].thr_id=getpid();
	
	_jkthread_kludge[getpid()] = &_jkthreadinfos[0];
	
	atexit( _jkthread_exit );
	
	signal( SIGINT, _jkthread_SIGINT );
	signal( SIGTERM, _jkthread_SIGTERM );	
	
	{

#if 0
		struct sigaction a;
		a.sa_handler = _jkthread_SIGCHLD;
		a.sa_mask = 0;
		a.sa_flags = SA_RESTART;
		
	    sigaction( SIGCHLD, &a, 0 );	
#else
		signal( SIGCHLD, _jkthread_SIGCHLD ); 
#endif
	}
	
	atexit( _jkthread_exit );

}

void _jkthread_exit(void)
{
	int i;
	
    /* destroy the _jksyssem */
	
	jksem_kill( &_jksyssem );
	
	/* destroy the _jklibc_sem */
	
	jksem_kill( &_jklibc_sem );
	
	/* destroy all other semaphores */
	
	for( i=0; i<JKMAX_SEMS; ++i )
	{
		if( _jksems[i].sem_id!=0 )
    	    jksem_kill( &_jksems[i] );	
	}
	
	/* kill all threads */
	
	for( i=1; i<JKMAX_THREADS; ++i )
	{
		if( _jkthreadinfos[i].thr_id!=0 )
    	    kill( _jkthreadinfos[i].thr_id, SIGTERM );
	}    		
}

void _jkthread_SIGINT(int s)
{
	DBG( ("pid %d got SIGINT\n", getpid() ) );	
    _jkthread_exit();	
	exit(0);
}

void _jkthread_SIGTERM(int s)
{
	DBG( ("pid %d got SIGTERM\n", getpid() ) );
    _jkthread_exit();	
	exit(0);
}

void _jkthread_SIGCHLD (int signum)
{
    int pid;
    int status;
	int i;
	jkthreadinfo *thr_info;

    while (1)
	{		
        pid = waitpid (WAIT_ANY, &status, WNOHANG);
        if (pid < 0)
	    {			
            break;
	    }
		
		
        if (pid == 0)
	        break;
		    
		/* a child exited - we need to reset the kludge array for that PID */


		  
		thr_info = _jkthread_kludge[pid];		
		_jkthread_kludge[ pid ] = &_jkdummythread;
		
		/* release any semaphores that were owned by him */
		
        for( i=0; i<JKMAX_SEMS; ++i )
		{
		    if( _jksems[i].sem_id!=0 && _jksems[i].owner==pid )
			{
				/* change ownership of the semphore to us */
				
			    _jksems[i].owner = getpid();
				
				/* now release it */
				
				jksem_release( &_jksems[i] );
			}
			
		}
		
		
		/* now call the termination function, if needed */
		
		if( thr_info->term_func )
		{
		    thr_info->term_func( thr_info->start_param );	

		}
		thr_info->thr_id=0;		
		   
        DBG( ( "%d: Child %d died\n", getpid(), pid ) );
	}
	
    signal( SIGCHLD, _jkthread_SIGCHLD );  
}

int _linux_thread_create(void (*fn)(void *), void *data, int stacksz )
{
	long retval;
	void **newstack;

	if( !_jkthread_inited )
	    jkthread_init();
	
	if (stacksz<JKMIN_STACK )
	    stacksz=JKMIN_STACK;

	/*
     * allocate new stack for subthread
	 */
    newstack = (void **) malloc(stacksz);
    if (!newstack)
	    return -1;
	
    /*
     * Set up the stack for child function, put the (void *)
	 * argument on the stack.
	 */
	
	newstack = (void **) (stacksz + (char *) newstack);
	*--newstack = data;
	
	/*
	 * Do clone() system call. We need to do the low-level stuff
	 * entirely in assembly as we're returning with a different
	 * stack in the child process and we couldn't otherwise guarantee
	 * that the program doesn't use the old stack incorrectly.
	 *
	 * Parameters to clone() system call:
	 *      %eax - __NR_clone, clone system call number
	 *      %ebx - clone_flags, bitmap of cloned data
	 *      %ecx - new stack pointer for cloned child
	 *
	 * In this example %ebx is CLONE_VM | CLONE_FS | CLONE_FILES |
	 * CLONE_SIGHAND which shares as much as possible between parent
	 * and child. (We or in the signal to be sent on child termination
	 * into clone_flags: SIGCHLD makes the cloned process work like
	 * a "normal" unix child process)
	 *
	 * The clone() system call returns (in %eax) the pid of the newly
	 * cloned process to the parent, and 0 to the cloned process. If
	 * an error occurs, the return value will be the negative errno.
	 *
	 * In the child process, we will do a "jsr" to the requested function
	 * and then do a "exit()" system call which will terminate the child.
	 */
	 __asm__ __volatile__(
	                      "int $0x80\n\t"         /* Linux/i386 system call */
			              "testl %0,%0\n\t"       /* check return value */
			              "jne 1f\n\t"            /* jump if parent */
			              "call *%3\n\t"          /* start subthread function */
			              "movl %2,%0\n\t"
			              "int $0x80\n"           /* exit system call: exit subthread */
			              "1:\t"
			             :"=a" (retval)
			             :"0" (__NR_clone),"i" (__NR_exit),
			              "r" (fn),
			              "b" (CLONE_VM | CLONE_FS | CLONE_FILES /*| CLONE_SIGHAND*/ | SIGCHLD),
			              "c" (newstack));
	
    if (retval < 0) 
	{
        errno = -retval;
		retval = -1;
	}
	
    return retval;
}



int jkthread_create(void (*fn)(void *), void *data, int stacksz, void (*term_fn)(void*))
{
    int ret;
	int i;
	jkthreadinfo *thread_info=0;

	
	if( !_jkthread_inited )
	{
	    jkthread_init();	
	}
	
	
	jksem_get( &_jksyssem );	

	for( i=1; i<JKMAX_THREADS; ++i )
	{
	    if( _jkthreadinfos[i].thr_id == 0 )
		{
		    thread_info = &_jkthreadinfos[i];	
			break;
		}		
	}
	
	if( !thread_info )
	{
		/* cant find one. Release the semaphore and fail */
		
		errno = ENOMEM;
		jksem_release( &_jksyssem );
		
		DBG( ("Can't find free thread space\n" ) );
		return -1;
	}
	
	thread_info->start_func = fn;
	thread_info->start_param = data;
	thread_info->term_func = term_fn;
	
	ret=_linux_thread_create(_jkthread_starter,thread_info,stacksz);

    if (ret>0 )
	{				
		/* update the thread info slot */
		
    	thread_info->thr_id = ret;
		thread_info->thr_errno = 0;
		thread_info->thr_h_errno = 0;
		  
		_jkthread_kludge[ ret ] = thread_info;
		
		DBG( ("In Parent, child=%d\n", ret ) );
	}
	
	
	jksem_release(&_jksyssem );
    return ret;		
}

void _jkthread_starter( void *p )
{	
	jkthreadinfo *i = (jkthreadinfo *)p;
	
	/* wait for original create thread call to finish updating kludge       */
	/* so errno and h_errno and thread local storage will work immediately  */
	/* in the meantime we are using the _jkdummythread space                */
		
	jksem_get( &_jksyssem );
	jksem_release( &_jksyssem );

	/* set up our signal handlers */
	
    signal( SIGCHLD, _jkthread_SIGCHLD ); 
	signal( SIGTERM, SIG_DFL );
	signal( SIGINT, SIG_DFL );
	
	DBG( ("In child %d, param =0x%08lx\n", getpid(), i->start_param ) );
		
	i->start_func(i->start_param);

	DBG( ("Child ended %d\n", getpid() ) );
}



void jkthread_kill( pid_t id )
{
	jksem_get( &_jksyssem );
    kill( id, SIGTERM );
	jksem_release( &_jksyssem );	
}

jksem *jksem_create( void )
{
	
    return jksem_create1( 1 );	
}


jksem *jksem_create1( int initial_count )
{
	jksem *s=0;
    int i;
	
	if( !_jkthread_inited )
	    jkthread_init();

	jksem_get( &_jksyssem );
	
	for( i=0; i<JKMAX_SEMS; ++i )
	{
	    if( _jksems[i].sem_id==0 )
		{
		    s=&_jksems[i];
			jksem_init1(s,initial_count);
			break;
		}
		
	}
    jksem_release( &_jksyssem );
	return s;
}


void jksem_init( jksem *s )
{
	jksem_init1( s, 1 );
}

				 
void jksem_init1( jksem *s, int initial_count )
{
	int		id;
	union semun	arg;

	
	if( !_jkthread_inited )
	{
	    jkthread_init();	
	}
	
	s->owner=0;
	s->usage=0;

	s->sem_id = semget( IPC_PRIVATE, 1, 0664 | IPC_CREAT | IPC_EXCL );
	if( s->sem_id == -1 )
	{
		perror( "Couldn't create semaphore" );
		exit(1);
	}
	
	arg.val = initial_count;
	if( semctl( s->sem_id, 0, SETVAL, arg ) == -1 )
	{
		perror( "Can't set semaphore value" );
        exit(1);
	}	
	
	
}




void jksem_kill( jksem *s )
{
	union semun	arg;
	
	semctl( s->sem_id, IPC_RMID, 0, arg );
    s->sem_id=0;	
}



int jksem_tryget( jksem *s )
{
	int ret;

    /* try get the semaphore */
  	struct sembuf	sops[1];

   	sops[0].sem_num = 0; 
  	sops[0].sem_op = -1;
   	sops[0].sem_flg = IPC_NOWAIT;

   	do 
   	{
   	    ret=semop(s->sem_id,sops,1);	
   	} while( ret==-1 && errno==EINTR );

	if( ret==-1 && errno==EAGAIN)
	{
	    return 0;
	}
    else
	{
        /* TODO: Is there a race condition here? */	
	    s->owner = getpid();
		s->usage = 1;
        return 1;		
	}
}


int jksem_get( jksem *s )
{
	int ret;

    /* try get the semaphore */
    
	struct sembuf	sops[1];
    sops[0].sem_num = 0; 
    sops[0].sem_op = -1;
    sops[0].sem_flg = 0;

    do 
    {
        ret=semop(s->sem_id,sops,1);	
    } while( ret==-1 && errno==EINTR );

	/* ok, we got the semaphore. We Own it now! */
    /* TODO: Is there a race condition here? */	
	s->owner = getpid();
	s->usage = 1;
	
	return 1;
}

		
		 


void jksem_release( jksem *s )
{
    int		id;
    struct sembuf	sops[1];
    int     ret;

	/* we dont own it anymore */
	
	s->owner=0;

    sops[0].sem_num = 0; 
    sops[0].sem_op = 1;
    sops[0].sem_flg = 0;
				
	do 
	{
	    ret=semop(s->sem_id,sops,1);	
	} while( ret==-1 && errno==EINTR );
	
}

		 
		
int jkrsem_tryget( jksem *s )
{
	int ret;

	if( s->owner == getpid() )
	{
		/* we already own it! */
		++s->usage;
	    return 1;	
	}
	else
	{		
		return jksem_tryget(s);
	}
}


int jkrsem_get( jksem *s )
{
	int ret;

	if( s->owner == getpid() )
	{
		/* we already own it! */
		++s->usage;
	    return 1;	
	}
	else
	{
		return jksem_get(s);
	}
	
}




void jkrsem_release( jksem *s )
{
	if( s->owner == getpid() )
	{
		/* we own it. decrease usage count */
		if( --s->usage == 0 )
    	    jksem_release(s);
	}	
}

